# coding=utf-8

from __future__ import unicode_literals

from .base import Base
from .generator import generator_of
from .other import other_obj
from .normal import normal_attr
from .urls import (
    COLLECTION_DETAIL_URL,
    COLLECTION_CONTENTS_URL,
    COLLECTION_COMMENTS_URL,
    COLLECTION_FOLLOWERS_URL,
)


class Collection(Base):
    def __init__(self, cid, cache, session):
        super(Collection, self).__init__(cid, cache, session)

    def _build_url(self):
        return COLLECTION_DETAIL_URL.format(self.id)

    # ---- simple info -----

    @property
    @normal_attr()
    def answer_count(self):
        return None

    @property
    @normal_attr()
    def created_time(self):
        return None

    @property
    @other_obj('people')
    def creator(self):
        return None

    @property
    @normal_attr()
    def comment_count(self):
        return None

    @property
    @normal_attr()
    def description(self):
        return None

    @property
    @normal_attr()
    def follower_count(self):
        return None

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr()
    def is_public(self):
        return None

    @property
    @normal_attr()
    def title(self):
        return None

    @property
    @normal_attr()
    def updated_time(self):
        return None

    # ----- generators -----

    @property
    def answers(self):
        """
        获取收藏夹里的所有答案。

        ..  warning::  无法被 shield

            因为内部是调用 :any:`Collection.contents` 的，
            所以此生成器无法被 :any:`shield` 保护。

            但是内部其实是用 shield 保护过 contents 的获取的，
            如果这个生成器异常了那还是处理下吧。

        ..  seealso:: :any:`Collection.articles`, :any:`Collection.contents`
        """
        from .answer import Answer
        from ..helpers import shield

        contents = self.contents
        if contents is None:
            return

        # noinspection PyTypeChecker
        for x in shield(contents):
            if isinstance(x, Answer):
                yield x

    @property
    def articles(self):
        """
        获取收藏夹里的所有文章。

        ..  warning::  无法被 shield

            因为内部是调用 :any:`Collection.contents` 的，
            所以此生成器无法被 :any:`shield` 保护。

            但是内部其实是用 shield 保护过 contents 的获取的，
            如果这个生成器异常了那还是处理下吧。

        ..  seealso:: :any:`Collection.answers`, :any:`Collection.contents`
        """
        from .article import Article
        from ..helpers import shield

        contents = self.contents
        if contents is None:
            return

        # noinspection PyTypeChecker
        for x in shield(contents):
            if isinstance(x, Article):
                yield x

    @property
    @generator_of(COLLECTION_COMMENTS_URL)
    def comments(self):
        return None

    @property
    @generator_of(COLLECTION_CONTENTS_URL, 'CollectionContent')
    def contents(self):
        """
        新版知乎专栏支持收藏文章了，这个生成器生成的对象可能是 :any:`Answer` 也可能是
        :any:`Article`，使用时要用 ``isinstance`` 判断类型后再获取对应对象的属性。

        ..  code-block:: python

            from zhihu_oauth import ZhihuClient, Answer, Article

            collection = client.collection(37770691)

            for content in collection.contents:
                if isinstance(content, Answer):
                    answer = content
                    print(answer.question.title)
                elif isinstance(content, Article):
                    article = content
                    print(article.title)

        如果你只需要答案或者只需要文章类型的数据，可以使用 :any:`Collection.answers`
        或者 :any:`Collection.articles` 进行获取。

        不过需要注意的是，这两个属性内部其实会调用 :any:`Collection.contents`，
        然后只返回相应类型的对象。所以其实也是遍历了所有内容的，
        效率与使用本函数然后自己判断类型一样。

        ..  seealso:: :any:`Collection.answers`, :any:`Collection.articles`
        """
        return None

    @property
    @generator_of(COLLECTION_FOLLOWERS_URL, 'people')
    def followers(self):
        """
        ..  warning:: 注意！

            知乎的这个 API 有问题，返回一些之后会将 is_end 设置为 True，
            导致无法获取到所有的关注者。

            并且此问题在知乎官方 Android APP 上也存在。你可以试着
            找个很多人关注的收藏夹，然后查看关注者，一直往下拉。
            大概加载 100 - 200（不固定，有时候一个都出不来）
            之后就没法往下刷了。

            起码在我这个地区是这样的。欢迎各路少侠反馈。

        """
        # TODO: collection.followers 这个 API 不稳定
        return None



# coding=utf-8

from __future__ import unicode_literals

from .base import Base
from .other import other_obj
from .normal import normal_attr

__all__ = ['Message']


class Message(Base):
    def __init__(self, mid, cache, session):
        super(Message, self).__init__(mid, cache, session)

    def _build_url(self):
        return ''

    # ----- simple info -----

    @property
    @normal_attr()
    def created_time(self):
        return None

    @property
    @normal_attr()
    def content(self):
        return None

    @property
    @other_obj('people')
    def sender(self):
        return None

    @property
    @other_obj('people')
    def receiver(self):
        return None

    def format(self, template='[{time}] {sender} --> {receiver}: {content}'):
        """
        格式化输出消息

        ``{time}`` 时间戳；``{sender}`` 发送者用户名；``{receiver}`` 接收者用户名；
        ``{content}`` 消息内容

        用法参见示例。

        :param str template: 格式化模板
        :return: 格式化后的字符串
        :rtype: str
        """
        return template.format(
            time=self.created_time,
            sender=self.sender.name,
            receiver=self.receiver.name,
            content=self.content,
        )

    def __str__(self):
        return self.format()

    __repr__ = __str__



# coding=utf-8

from __future__ import unicode_literals

import datetime
import warnings

import requests.packages.urllib3 as urllib3

from .zhcls.activity import Activity, ActType
from .zhcls.generator import BaseGenerator, ActivityGenerator
from .zhcls.streaming import StreamingJSON
from .zhcls.utils import SimpleEnum
from .exception import ZhihuException, ZhihuWarning


__all__ = ['ActivityFormatter', 'SHIELD_ACTION', 'act2str', 'shield', 'ts2str']

SHIELD_ACTION = SimpleEnum(
    ['EXCEPTION', 'PASS', 'STOP']
)
"""
ActType 是用于表示 shield 抵挡 Exception 达到最大次数后的动作的枚举类，取值如下：

================= ====================
常量名              说明
================= ====================
EXCEPTION          抛出异常
PASS               跳过，获取下一个数据
STOP               结束处理
================= ====================
"""


def shield(inner, durability=3, start_at=0, action=SHIELD_ACTION.EXCEPTION):
    """
    shield 函数用于自动处理知乎的各种生成器
    （如 :any:`People.followers`, :any:`Question.answers`） 在获取分页数据时出错的情况。

    ..  warning:: 用户动态的生成器因为获取方式比较特殊，无法被 shield 保护

    用法：

    比如我们想获取关注了某个专栏的用户分别关注了哪些话题……

    ..  code-block:: python

        column = client.column('zijingnotes')
        result = []
        for user in shield(column.followers, action=SHIELD_ACTION.PASS):
            L = []
            print('Start proc user', user.name)
            if user.over:
                print(user.over_reason)
                continue
            for topic in shield(user.following_topics):
                print('Add topic', topic.name)
                L.append(topic.name)
            result.append(L)

        # output result

    :param inner: 需要被保护的生成器
    :param int durability: 耐久度，表示获取同一数据最多连续出错几次
    :param int start_at: 从第几个数据开始获取
    :param action: 当耐久度消耗完后的动作，参见 :any:`SHIELD_ACTION`，默认动作是抛出异常
    :return: 新的生成器……
    """
    if not isinstance(inner, BaseGenerator):
        raise ValueError('First argument must be Zhihu Generator Classes')
    if isinstance(inner, ActivityGenerator):
        raise ValueError(' Activity Generator is the only one can\'t be shield')
    offset = start_at
    hp = durability
    while True:
        i = -1
        try:
            for i, x in enumerate(inner.jump(offset)):
                yield x
                hp = durability
            break
        except (ZhihuException, urllib3.exceptions.MaxRetryError) as e:
            offset += i + 1
            hp -= 1
            warnings.warn(
                '[{type}: {e}] be shield when get NO.{offset} data'.format(
                    type=e.__class__.__name__,
                    e=e,
                    offset=offset
                ),
                ZhihuWarning
            )
            if hp == 0:
                if action is SHIELD_ACTION.EXCEPTION:
                    raise e
                elif action is SHIELD_ACTION.PASS:
                    offset += 1
                    hp = durability
                elif action is SHIELD_ACTION.STOP:
                    break
                else:
                    raise e


def ts2str(ts, fmt=None, offset=8):
    """
    将时间戳转换为表示时间的字符串。

    :param int ts: 精确到秒的 unix timestamp
    :param fmt: 格式化文本，默认值 ``%Y-%m-%d %H:%M:%S``
    :param offset: 当前时区偏移，单位小时，默认为 8 小时
    :return: 时间戳转换为时间的字符串表示
    :rtype: str
    """
    offset *= 3600    # 时区偏移
    if fmt is None:
        fmt = '%Y-%m-%d %H:%M:%S'
    return datetime.datetime.utcfromtimestamp(ts + offset).strftime(fmt)


_DEFAULT_ACTIVITY_FORMATTER_MANY_ONE = [
    ({
        ActType.CREATE_ARTICLE, ActType.CREATE_QUESTION,
        ActType.FOLLOW_COLLECTION, ActType.FOLLOW_COLUMN,
        ActType.FOLLOW_QUESTION, ActType.JOIN_LIVE, ActType.PUBLISH_LIVE,
        ActType.VOTEUP_ARTICLE, ActType.VOTEUP_EBOOK
    }, '{act.action_text} 「{act.target.title}」'),
    ({
        ActType.FOLLOW_ROUNDTABLE,
        ActType.FOLLOW_TOPIC
    }, '{act.action_text} 「{act.target.name}」'),
]

_DEFAULT_ACTIVITY_FORMATTER_ONE_ONE = {
    ActType.CREATE_ANSWER: '{act.action_text} 「{act.target.question.title}」',
    ActType.VOTEUP_ANSWER: '{act.action_text} 「{act.target.question.title}」'
                           ' by 「{act.target.author.name}」',
    ActType.CREATE_PIN: '{act.action_text} {act.target.excerpt_title}',
    ActType.LIKE_PIN: '{act.action_text} {act.target.excerpt_title}'
                      ' by 「{act.target.author.name}」',
    ActType.COLLECT_ANSWER: '{act.action_text} '
                            '「{act.target[answer].question.title}」'
                            ' by 「{act.target[answer].author.name}」'
                            ' 到收藏夹 「{act.target[collection].title}」',
    ActType.COLLECT_ARTICLE: '{act.action_text} 「{act.target[article].title}」'
                             ' 到收藏夹 「{act.target[collection].title}」'
}


class ActivityFormatter(object):
    """
    这是将 Activity 转换为字符串的辅助类，一般情况下不需要使用，直接使用辅助函数
    :any:`act2str` 即可。

    如果你需要自定义格式化模板，请参考下面的用法：

    ..  code-block:: python

        class MyActivityFormatter(ActivityFormatter):
            def __init__(self, user_name):
                self._user_name = user_name

            def like_pin_formatter(self, act):
                content_summary = next(act.target.contents).content[:20]
                return '{i} 赞了 {act.target.author.name} 的分享： {content}'.format(
                    i=self._user_name, act=act, content=content_summary,
                )

            create_pin_formatter = '{act.action_text} 一些东西'

        guxizhao = client.people('guxizhao')

        formatter = MyActivityFormatter(guxizhao.name)

        for act in guxizhao.activities:
            print(ts2str(act.created_time), formatter(act))


    ..  note:: 执行结果

        除了发表分享和对分享点赞这两个类型的 Activity 之外，其他类型的格式化结果均和
        :any:`act2str` 函数一致。

        ActType.LIKE_PIN 类型的会被转换成 ``xxx 赞同了 yyy 的分享：<分享内容的前20字>``

        ActType.CREATE_PIN 类型的会被转换成 ``xxx 分享了 一些东西``

    简单来说就是你可以继承 :any:`ActivityFormatter` 类，然后定义一些函数或者常量，名称是
    ActType 类型的小写形式 + ``_formatter``。

    如果它是个函数，需要接受一个 :any:`ActType` 或者 :any:`StreamingJSON` 对象，
    返回一个字符串模板。如果直接是个变量那就直接被当成模板使用。

    模板里一律用 ``act`` 代表 :any:`Activity` 对象。
    """
    @staticmethod
    def __check_type(act):
        return isinstance(act, Activity) \
               or (
                   isinstance(act, StreamingJSON)
                   and hasattr(act, 'type')
                   and act.type in ActType
               )

    def __call__(self, act):
        if not self.__check_type(act):
            raise ValueError('Only support Activity objects.')

        attr_name = act.type.lower() + '_formatter'

        if hasattr(self, attr_name):
            fmt = getattr(self, attr_name)
            if hasattr(fmt, '__call__'):
                fmt = fmt(act)
            assert isinstance(fmt, str), \
                'Formatter must be a str of a function like ' \
                'func(act: Activity) -> str'
        elif act.type in _DEFAULT_ACTIVITY_FORMATTER_ONE_ONE:
            fmt = _DEFAULT_ACTIVITY_FORMATTER_ONE_ONE[act.type]
        else:
            found = False
            for ks, v in _DEFAULT_ACTIVITY_FORMATTER_MANY_ONE:
                if act.type in ks:
                    found = True
                    fmt = v
            assert found, 'Unknown ActType ' + act.type

        # noinspection PyUnboundLocalVariable
        return fmt.format(act=act)

act2str = ActivityFormatter()
"""
一个将 ``xxx.activities`` 返回的 :any:`Activity` 或 :any:`StreamingJSON`
对象转换为可读的字符串的辅助函数。例子：

..  code-block:: python

    guxizhao = client.people('guxizhao')

    for act in guxizhao.activities:
        print(ts2str(act.created_time), act2str(act))

..  note:: 结果

    2016-12-08 01:18:20 顾惜朝关注了问题
    「如何评价美剧《西部世界》（Westworld）第一季第十集（S01E10）?」

    2016-12-07 23:18:02 顾惜朝赞同了回答
    「如何看待民生银行性骚扰事件？」 by 「meta」

    2016-12-07 13:20:32 顾惜朝赞了文章
    「冒牌高校教授，正版绝命毒师」

    2016-12-07 10:59:43 顾惜朝赞同了回答
    「你最喜欢的故事是什么？」 by 「胡不归」

    2016-12-07 01:30:51 顾惜朝赞了文章
    「专访 | 关于《西部世界》季终集，死过一千次的泰迪这样说」

    2016-12-06 20:23:59 顾惜朝赞同了回答
    「如何看待知乎上「一知半解」也要强行回答的回答者？」 by 「河森堡」

    2016-12-06 01:57:30 顾惜朝关注了收藏夹 「你丫竟然在图片里下毒？！！」
    2016-12-05 22:23:25 顾惜朝赞同了回答
    「《你的名字。》中彗星碎片撞击地面的影响是否合理？」 by 「Macro kuo」

    2016-12-04 00:36:24 顾惜朝赞了文章 「Google Earth: 这才是真·情怀」

如果想自定义描述模板，请参见 :any:`ActivityFormatter` 类。
"""



# coding=utf-8

from __future__ import unicode_literals

import itertools
import warnings

from ..exception import (
    MyJSONDecodeError,
    UnexpectedResponseException,
    CantGetTickets
)
from .base import Base
from .generator import generator_of
from .normal import normal_attr
from .other import other_obj
from .streaming import streaming
from .urls import (
    LIVE_DETAIL_URL,
    LIVE_ENDED_URL,
    LIVE_MEMBERS_URL,
    LIVE_ONGOING_URL,
    LIVE_RELATED_URL,
    LIVE_TICKETS_URL,
    LIVE_TICKETS_ENDED_URL,
    LIVE_TICKETS_QUIET_URL,
)

__all__ = ['Live', 'LiveTag', 'LiveTicket', 'LiveBadge']


class LiveBadge(Base):
    def __init__(self, lbid, cache, session):
        super(LiveBadge, self).__init__(lbid, cache, session)

    def _build_url(self):
        return None

    @property
    @normal_attr()
    def avatar_url(self):
        return None

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr()
    def name(self):
        return None


class LiveTag(Base):
    def __init__(self, ltid, cache, session):
        super(LiveTag, self).__init__(ltid, cache, session)

    def _build_url(self):
        return None

    @property
    @normal_attr('available_num')
    def available_count(self):
        return None

    @property
    @normal_attr()
    def created_at(self):
        return None

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr('live_num')
    def live_count(self):
        return None

    @property
    @normal_attr()
    def name(self):
        return None

    @property
    @normal_attr()
    def score(self):
        return None

    # ----- generators -----

    @property
    @generator_of(LIVE_ONGOING_URL, 'LiveOfTag')
    def lives_ongoing(self):
        return None

    @property
    @generator_of(LIVE_ENDED_URL, 'LiveOfTag')
    def lives_ended(self):
        return None

    @property
    def lives(self):
        from ..helpers import shield
        for live in itertools.chain(
                shield(self.lives_ongoing),
                shield(self.lives_ended)
        ):
            yield live


class LiveTicket(Base):
    def __init__(self, product_id, cache, session):
        super(LiveTicket, self).__init__(product_id, cache, session)

    def _build_url(self):
        return None

    @property
    @other_obj('LiveBadge', module_filename='live')
    def badge(self):
        return None

    @property
    @normal_attr('product_id')
    def id(self):
        return self._id

    @property
    @streaming('price')
    def __price(self):
        return None

    @property
    def price(self):
        return self.__price.amount

    @property
    def price_unit(self):
        return self.__price.unit


class Live(Base):
    def __init__(self, lid, cache, session):
        super(Live, self).__init__(lid, cache, session)

    def _build_url(self):
        return LIVE_DETAIL_URL.format(self.id)

    # ----- simple info -----

    @property
    @normal_attr()
    def alert(self):
        """
        提示语，就是客户端里显示为淡蓝色背景的那一块文字。
        """
        return None

    @property
    @normal_attr()
    def can_speak(self):
        return None

    @property
    @normal_attr()
    def created_at(self):
        return None

    @property
    @normal_attr()
    def description(self):
        return None

    @property
    @normal_attr()
    def ends_at(self):
        return None

    @property
    @normal_attr()
    def ends_in(self):
        """
        正数表示还剩多久结束，0 应该表示已经结束了，如果是负数表示…………表示啥呢？
        """
        return None

    @property
    @streaming('fee')
    def __fee(self):
        return None

    @property
    def fee(self):
        """
        费用（一般这是最低票价），不过数值是 x 100 的，比如 999 表示 9.99
        """
        return self.__fee.amount

    @property
    def fee_unit(self):
        """
        费用的单位，一般就是 RMB 吧…………
        """
        return self.__fee.unit

    @property
    @normal_attr()
    def feedback_score(self):
        """
        反馈评分，应该是 0 - 5 吧。
        """
        return None

    @property
    @normal_attr()
    def has_feedback(self):
        """
        是否有反馈？
        """
        return None

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr()
    def is_admin(self):
        return None

    @property
    @normal_attr()
    def in_promotion(self):
        """
        是否处于促销中
        """
        return None

    @property
    @normal_attr()
    def is_muted(self):
        return None

    @property
    @normal_attr()
    def liked(self):
        return None

    @property
    def liked_count(self):
        return self.liked_num

    @property
    @normal_attr()
    def liked_num(self):
        return None

    @property
    @normal_attr()
    def note(self):
        return None

    @property
    @normal_attr()
    def purchasable(self):
        """
        可否购买？
        """
        return None

    @property
    @normal_attr()
    def role(self):
        """
        返回一个字符串，表示于 Live 的关系。

        'visitor' 表示未参与 Live。

        ‘audience’ 表示参与了 Live，作为观众。

        '<一个我不知道的值>' 表示是组织者，因为我没开过 Live，所以不知道是什么值。

        '<另一个我不知道的值>' 表示是协作者，我也没协助过别人，所以也不知道 =。=
        """
        return None

    @property
    @streaming('seats')
    def seat(self):
        """
        Live 参与情况

        常见返回值：

        ..  code-block: javascript

            {
                "max": 500,     // 最多 500 人参与
                "taken": 278,   // 已有 278 人参与
            }

        做了两个 shortcut 属性 `:any:`seat_max` 和 :any:`seat_taken`， 可以直接使用。
        """
        return None

    @property
    def seat_max(self):
        """
        最大参与人数，其实是从 :any:`seat` 属性里取的。
        """
        return self.seat.max

    @property
    def seat_taken(self):
        """
        已参与人数，其实是从 :any:`seat` 属性里取的。
        """
        return self.seat.taken

    @property
    @streaming('speaker')
    def __speaker(self):
        return None

    @property
    @other_obj('people', 'READ_FROM_RETURN_VALUE')
    def speaker(self):
        """
        演讲者，:any:`People` 对象。
        """
        return self.__speaker.member.raw_data()

    @property
    @normal_attr()
    def starts_at(self):
        return None

    @property
    @normal_attr()
    def subject(self):
        """
        Live 的主题，其实就是标题，所以有同功能的属性 :any:`Live.title`
        """
        return None

    @property
    def title(self):
        """
        ..  seealso:: :any:`subject`
        """
        return self.subject

    # ----- generators -----

    @property
    @streaming('cospeakers')
    def __cospeakers(self):
        return None

    @property
    def cospeakers(self):
        from .people import People

        # noinspection PyTypeChecker
        for people in self.__cospeakers:
            yield People(people.id, people.raw_data(), self._session)

    @property
    @generator_of(LIVE_MEMBERS_URL, 'PeopleWithLiveBadge')
    def participants(self):
        """
        参与 Live 的人，这个生成器用法比较奇特，请看下面的例子：

        ..  code-block:: python

            live = client.live(789426202925346816)

            for role, badge, people in live.members:
                print(role, badge.name, people.name)

        其中 role 为 'audience' 表示观众，除了这个值暂时没发现别的取值。

        badge 为一个 :any:`LiveBadge` 对象，一般能用到的也就 id 和 name 属性。

        第三个 people 就是标准的 :any:`People` 对象了。
        """
        return None

    @property
    @generator_of(LIVE_RELATED_URL, 'live')
    def related(self):
        return None

    @property
    @streaming('tags')
    def __tags(self):
        return None

    @property
    def tags(self):
        """
        返回 :any:`LiveTag` 对象的生成器，但目前看来应该每个 Live 只有一个 Tag。
        """
        # noinspection PyTypeChecker
        for tag in self.__tags:
            yield LiveTag(tag.id, tag.raw_data(), self._session)

    def __try_ticket_request(self, url):
        res = self._session.post(url)
        try:
            data = res.json()
            return data
        except MyJSONDecodeError:
            raise UnexpectedResponseException(
                url, res,
                'a json string.',
            )

    @property
    def tickets(self):
        """
        ..  warning::

            此接口无法用于当前登录用户已参与的 Live。当强行调用时，
            此函数将产生一个警告并且不会有任何返回。

        正常情况下返回的是 :any:`LiveTicket` 对象的生成器。
        """

        if self.role != 'visitor':
            warnings.warn(CantGetTickets)
        else:
            normal_url = LIVE_TICKETS_URL.format(self.id)
            quiet_url = LIVE_TICKETS_QUIET_URL.format(self.id)
            ended_url = LIVE_TICKETS_ENDED_URL.format(self.id)

            if self.ends_in == 0:
                url = ended_url
            elif self.seat_taken >= self.seat_max:
                url = quiet_url
            else:
                url = normal_url

            data = self.__try_ticket_request(url)

            if 'error' in data and 'code' in data['error']:
                if data['error']['code'] == 4046:
                    data = self.__try_ticket_request(quiet_url)
                elif data['error']['code'] == 4048:
                    data = self.__try_ticket_request(ended_url)

            try:
                data = data['product_list']
            except KeyError:
                raise UnexpectedResponseException(
                    quiet_url, data,
                    'a json string contains [product_list] attr.',
                )

            for ticket in data:
                yield LiveTicket(ticket['product_id'], ticket, self._session)



# coding=utf-8

from __future__ import unicode_literals

import re

# ------- Zhihu API URLs --------

ZHIHU_API_ROOT = 'https://api.zhihu.com'

# ----- 用户相关 -----

# self - GET - 获取自身资料

SELF_DETAIL_URL = ZHIHU_API_ROOT + '/people/self'

# people - GET - 详情

PEOPLE_DETAIL_URL = ZHIHU_API_ROOT + '/people/{}'

# people.answers - GET - 回答

PEOPLE_ANSWERS_URL = PEOPLE_DETAIL_URL + '/answers'

# people.articles - GET - 文章

PEOPLE_ARTICLES_URL = PEOPLE_DETAIL_URL + '/articles'

# people.collections - GET - 收藏夹

PEOPLE_COLLECTIONS_URL = PEOPLE_DETAIL_URL + '/collections_v2'

# people.columns - GET - 专栏

PEOPLE_COLUMNS_URL = PEOPLE_DETAIL_URL + '/columns'

# people.followers - GET - 粉丝
# me.follow - POST - 关注用户

PEOPLE_FOLLOWERS_URL = PEOPLE_DETAIL_URL + '/followers'

# me.follow - DELETE - 取消关注用户

PEOPLE_CANCEL_FOLLOWERS_URL = PEOPLE_FOLLOWERS_URL + '/{}'

# me.following_collections - GET - 关注的收藏夹

PEOPLE_FOLLOWING_COLLECTIONS_URL = PEOPLE_DETAIL_URL + '/following_collections'

# people.following_columns - GET - 关注的专栏

PEOPLE_FOLLOWING_COLUMNS_URL = PEOPLE_DETAIL_URL + '/following_columns'

# people.following_questions - GET - 关注的问题

PEOPLE_FOLLOWING_QUESTIONS_URL = PEOPLE_DETAIL_URL + '/following_questions'

# people.following_topics - GET - 关注的话题

PEOPLE_FOLLOWING_TOPICS_URL = PEOPLE_DETAIL_URL + '/following_topics'

# people.followings - GET - 关注的人

PEOPLE_FOLLOWINGS_URL = PEOPLE_DETAIL_URL + '/followees'

# people.questions - GET - 用户提的问题

PEOPLE_QUESTIONS_URL = PEOPLE_DETAIL_URL + '/questions'

# people.activities - GET - 用户最近动态

PEOPLE_ACTIVITIES_URL = PEOPLE_DETAIL_URL + '/activities'

# people.lives - GET - 用户 Live（包括参与的和组织的）

PEOPLE_LIVES_URL = PEOPLE_DETAIL_URL + '/lives'

# people.liked_lives - GET - 用户感兴趣的 Live

PEOPLE_LIKED_LIVES_URL = ZHIHU_API_ROOT + '/lives/people/{}/like_lives'

# people.pins - GET - 用户的分享

PEOPLE_PINS_URL = PEOPLE_DETAIL_URL + '/pins'

# ----- 答案相关 -----

# answer - GET - 详情
# me.delete - DELETE - 删除答案

ANSWER_DETAIL_URL = ZHIHU_API_ROOT + '/answers/{}'

# answer.collections - GET - 所在收藏夹

ANSWER_COLLECTIONS_URL = ANSWER_DETAIL_URL + '/collections'

# me.collect - PUT - 加入收藏夹

ANSWER_COLLECT_URL = ANSWER_DETAIL_URL + '/collections_v2'

# answer.comment - GET - 评论

ANSWER_COMMENTS_URL = ANSWER_DETAIL_URL + '/comments'

# answer.voters - GET - 点赞用户
# me.vote - POST - 给答案投票

ANSWER_VOTERS_URL = ANSWER_DETAIL_URL + '/voters'

# me.thanks - POST - 给答案点感谢

ANSWER_THANKS_URL = ANSWER_DETAIL_URL + '/thankers'

# me.thanks - DELETE - 取消感谢

ANSWER_CANCEL_THANKS_URL = ANSWER_THANKS_URL + '/{}'

# me.unhelpful - POST - 没有帮助

ANSWER_UNHELPFUL_URL = ANSWER_DETAIL_URL + '/nothelpers'

# me.unhelpful - DELETE - 取消没有帮助

ANSWER_CANCEL_UNHELPFUL_URL = ANSWER_UNHELPFUL_URL + '/{}'

# ----- 问题相关 -----

# question - GET - 详情

QUESTION_DETAIL_URL = ZHIHU_API_ROOT + '/questions/{}'

# question.answers - GET - 回答

QUESTION_ANSWERS_URL = QUESTION_DETAIL_URL + '/answers'

# question.comments - GET - 评论

QUESTION_COMMENTS_URL = QUESTION_DETAIL_URL + '/comments'

# question.answers - GET - 关注者
# me.follow - POST - 关注问题

QUESTION_FOLLOWERS_URL = QUESTION_DETAIL_URL + '/followers'

# me.follower - DELETE - 取消关注
QUESTION_CANCEL_FOLLOWERS_URL = QUESTION_FOLLOWERS_URL + '/{}'

# question.topics - GET - 所属话题

QUESTION_TOPICS_URL = QUESTION_DETAIL_URL + '/topics'

# ----- 话题相关 -----

# topic - GET - 详情

TOPIC_DETAIL_URL = ZHIHU_API_ROOT + '/topics/{}'

# topic.activities - GET - 动态

TOPIC_ACTIVITIES_URL = TOPIC_DETAIL_URL + '/activities_new'

# topic.best_answers - GET - 精华回答

TOPIC_BEST_ANSWERS_URL = TOPIC_DETAIL_URL + '/best_answers'

# topic.best_answerers - GET - 最佳回答者

TOPIC_BEST_ANSWERERS_URL = TOPIC_DETAIL_URL + '/best_answerers'

# topic.children - GET - 子话题

TOPIC_CHILDREN_URL = TOPIC_DETAIL_URL + '/children'

# topic.children - GET - 父话题

TOPIC_PARENTS_URL = TOPIC_DETAIL_URL + '/parent'

# topic.unanswered_questions - GET - 未回答的问题

TOPIC_UNANSWERED_QUESTION = TOPIC_DETAIL_URL + '/unanswered_questions'

# topic.followers - GET - 关注者
# me.follow - POST - 关注话题

TOPIC_FOLLOWERS_URL = TOPIC_DETAIL_URL + '/followers'

# me.follow - DELETE - 取消关注

TOPIC_CANCEL_FOLLOW_URL = TOPIC_FOLLOWERS_URL + '/{}'

# ----- 收藏夹相关 -----

# collection - GET - 详情
# me.delete - DELETE - 删除收藏夹

COLLECTION_DETAIL_URL = ZHIHU_API_ROOT + '/collections/{}'

# collection.contents - GET - 所有收藏的内容（包括答案和文章）

COLLECTION_CONTENTS_URL = COLLECTION_DETAIL_URL + '/contents'

# collection.comments - GET - 评论

COLLECTION_COMMENTS_URL = COLLECTION_DETAIL_URL + '/comments'

# collection.followers - GET - 粉丝
# me.follow - POST - 关注专栏

COLLECTION_FOLLOWERS_URL = COLLECTION_DETAIL_URL + '/followers'

# me.follow - DELETE - 取消关注

COLLECTION_CANCEL_FOLLOW_URL = COLLECTION_FOLLOWERS_URL + '/{}'

# ----- 专栏相关 -----

# column - GET - 详情

COLUMN_DETAIL_URL = ZHIHU_API_ROOT + '/columns/{}'

# column.articles - GET - 文章

COLUMN_ARTICLES_URL = COLUMN_DETAIL_URL + '/articles'

# column.followers - GET - 关注者
# me.follow - POST - 关注专栏

COLUMN_FOLLOWERS_URL = COLUMN_DETAIL_URL + '/followers'

# me.follow - DELETE - 取消关注

COLUMN_CANCEL_FOLLOW_URL = COLUMN_FOLLOWERS_URL + '/{}'

# ----- 文章相关 -----

# article - GET - 详情
# me.delete - DELETE - 删除文章

ARTICLE_DETAIL_URL = ZHIHU_API_ROOT + '/articles/{}'

# article.vote - GET:  - 获取点赞用户（无效）
# me.vote - POST - 点赞

ARTICLE_VOTE_URL = ARTICLE_DETAIL_URL + '/voters'

# article.comments - GET - 评论

ARTICLE_COMMENTS_URL = ARTICLE_DETAIL_URL + '/comments'

# me.collect - PUT - 收藏

ARTICLE_COLLECT_URL = ARTICLE_DETAIL_URL + '/collections'

# ----- 评论相关 -----

# me.comment - POST - 发表评论

SEND_COMMENT_URL = ZHIHU_API_ROOT + '/comments'

# me.delete - DELETE - 删除评论

COMMENT_DETAIL_URL = ZHIHU_API_ROOT + '/comments/{}'

# comment.replies - GET - 评论的回复

COMMENT_REPLIES_URL = COMMENT_DETAIL_URL + '/replies'

# comment.conversation - GET - 评论的对话

COMMENT_CONVERSION_URL = COMMENT_DETAIL_URL + '/conversation'

# me.vote - POST - 给评论点赞

COMMENT_VOTE_URL = COMMENT_DETAIL_URL + '/voters'

# me.vote - DELETE - 取消点赞

COMMENT_CANCEL_VOTE_URL = COMMENT_VOTE_URL + '/{}'

# ----- Pin 分享相关 -----

# pin - GET - 详情

PIN_DETAIL_URL = ZHIHU_API_ROOT + '/pins/{}'

# pin.comments - GET - 评论

PIN_COMMENTS_URL = PIN_DETAIL_URL + '/comments'

# pin.voters - GET - 点赞者

PIN_VOTERS_URL = PIN_DETAIL_URL + '/likers'

# ----- Live 相关 -----

# live - GET - 详情

LIVE_DETAIL_URL = ZHIHU_API_ROOT + '/lives/{}'

# live.participants - GET - Live 参与者
# 后两个 API 只会给出是好友的参与者和不是好友的参与者，目前均未使用，因为含义不是很清楚

LIVE_MEMBERS_URL = LIVE_DETAIL_URL + '/members'
LIVE_MEMBERS_FRIENDS_URL = LIVE_DETAIL_URL + '/members/friends'
LIVE_MEMBERS_NON_FRIENDS_URL = LIVE_DETAIL_URL + '/members/nonfriends'

# live.related - GET - 相关 Live

LIVE_RELATED_URL = LIVE_DETAIL_URL + '/related'

# client.lives_ongoing - GET - 所有正在开放的 Live
# LiveTag.lives_ongoing - GET - 所有正在开放的 Live (需要附加 query 参数 tags = <tagid>）

LIVE_ONGOING_URL = ZHIHU_API_ROOT + '/lives/ongoing'

# client.lives_ended - GET - 所有已结束的 Live
# LiveTag.lives_ended - GET - 所有已结束的 Live (需要附加 query 参数 tags = <tagid>）

LIVE_ENDED_URL = ZHIHU_API_ROOT + '/lives/ended'

# client.lives_ended - GET - 所有已结束的 Live
# LiveTag.lives_ended - GET - 所有已结束的 Live (需要附加 query 参数 tags = <tagid>）

LIVE_TAGS_URL = ZHIHU_API_ROOT + '/lives/tags'

# live.tickets - GET - Live 票价

LIVE_TICKETS_URL = LIVE_DETAIL_URL + '/apply'
LIVE_TICKETS_QUIET_URL = LIVE_TICKETS_URL + '/quiet'    # 座位已满时的票价接口
LIVE_TICKETS_ENDED_URL = LIVE_TICKETS_URL + '/ended'    # Live 已结束的票价接口

# me.follow - POST - 感兴趣 Live

LIVE_LIKE_URL = LIVE_DETAIL_URL + '/like'

# ----- 私信相关 -----

# me.whispers - GET - 获取用户私信对话列表

WHISPERS_URL = ZHIHU_API_ROOT + '/inbox'

# whisper.messages - GET - 获取用户某一对话的消息列表

MESSAGES_URL = ZHIHU_API_ROOT + '/messages'

# me.message - POST - 发送私信

SEND_MESSAGE_URL = ZHIHU_API_ROOT + '/messages'

# ----- 搜索 -----

SEARCH_API_URL = ZHIHU_API_ROOT + '/search_v3'

# ----- 其他操作 -----

# me.block - POST - 屏蔽用户

BLOCK_PEOPLE_URL = ZHIHU_API_ROOT + '/settings/blocked_users'

# me.block - DELETE - 取消屏蔽用户

CANCEL_BLOCK_PEOPLE_URL = BLOCK_PEOPLE_URL + '/{}'

##############################
# ----- Zhihu Web URLs ----- #
##############################

ZHIHU_WEB_ROOT = 'https://www.zhihu.com'

re_answer_url = re.compile(
    r'^(?:https?://)?www.zhihu.com/question/\d+/answer/(\d+)/?$')
"""
答案 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_article_url = re.compile(r'^(?:https?://)?zhuanlan.zhihu.com/p/(\d+)/?$')
"""
文章 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_collection_url = re.compile(
    r'^(?:https?://)?www.zhihu.com/collection/(\d+)/?$')
"""
收藏夹 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_column_url = re.compile(r'^(?:https?://)?zhuanlan.zhihu.com/([^/ ]+)/?$')
"""
专栏 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_live_url = re.compile(r'^(?:https?://)?www.zhihu.com/lives/(\d+)/?$')
"""
Live URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_people_url = re.compile(r'^(?:https?://)?www.zhihu.com/people/([^/ ]+)/?$')
"""
用户 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_pin_url = re.compile(r'^(?:https?://)?www.zhihu.com/pin/(\d+)/?$')
"""
Pin URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_question_url = re.compile(r'^(?:https?://)?www.zhihu.com/question/(\d+)/?$')
"""
问题 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""

re_topic_url = re.compile(r'^(?:https?://)?www.zhihu.com/topic/(\d+)/?$')
"""
问题 URL 的正则，用于 :any:`ZhihuClient.from_url` 方法。
"""


RE_TYPE_MAP = {
    # RE             type      int id
    re_answer_url: ('answer', True),
    re_article_url: ('article', True),
    re_collection_url: ('collection', True),
    re_column_url: ('column', False),
    re_live_url: ('live', True),
    re_pin_url: ('pin', True),
    re_people_url: ('people', False),
    re_question_url: ('question', True),
    re_topic_url: ('topic', True),
}
"""
正则表达式于 :any:`ZhihuClient` 的方法的对应关系。

键是正则，值是二元组，两个值分别是方法名和是否需要将 ``id`` 转化为整数。
"""



# coding=utf-8

from __future__ import unicode_literals

import importlib

from .normal import normal_attr
from .streaming import StreamingJSON
from .utils import SimpleEnum
from ..exception import UnimplementedException


__all__ = ['Activity', 'ActType']


_verb_to_type_map = {
    'ANSWER_CREATE': 'CREATE_ANSWER',
    'ANSWER_VOTE_UP': 'VOTEUP_ANSWER',
    'EBOOK_VOTE_UP': 'VOTEUP_EBOOK',
    'LIVE_JOIN': 'JOIN_LIVE',
    'LIVE_PUBLISH': 'PUBLISH_LIVE',
    'MEMBER_COLLECT_ANSWER': 'COLLECT_ANSWER',
    'MEMBER_COLLECT_ARTICLE': 'COLLECT_ARTICLE',
    'MEMBER_CREATE_ARTICLE': 'CREATE_ARTICLE',
    'MEMBER_CREATE_PIN': 'CREATE_PIN',
    'MEMBER_FOLLOW_COLLECTION': 'FOLLOW_COLLECTION',
    'MEMBER_FOLLOW_COLUMN': 'FOLLOW_COLUMN',
    'MEMBER_FOLLOW_ROUNDTABLE': 'FOLLOW_ROUNDTABLE',
    'MEMBER_FOLLOW_TOPIC': 'FOLLOW_TOPIC',
    'MEMBER_LIKE_PIN': 'LIKE_PIN',
    'MEMBER_VOTEUP_ARTICLE': 'VOTEUP_ARTICLE',
    'QUESTION_CREATE': 'CREATE_QUESTION',
    'QUESTION_FOLLOW': 'FOLLOW_QUESTION',
    'TOPIC_FOLLOW': 'FOLLOW_TOPIC',
}

ActType = SimpleEnum(_verb_to_type_map.values())
"""
ActType 是用于表示用户动态类型的枚举类，可供使用的常量有：

================= ================ ======================
常量名              说明              `target` 属性类型
================= ================ ======================
COLLECT_ANSWER     收藏答案          比较特殊，见下文
COLLECT_ARTICLE    收藏文章          比较特殊，见下文
CREATE_ANSWER      回答问题          :any:`Answer`
CREATE_ARTICLE     发表文章          :any:`Article`
CREATE_PIN         发表分享          :any:`StreamingJSON`
CREATE_QUESTION    提出问题          :any:`Question`
FOLLOW_COLLECTION  关注收藏夹        :any:`Collection`
FOLLOW_COLUMN      关注专栏          :any:`Column`
FOLLOW_QUESTION    关注问题          :any:`Question`
FOLLOW_ROUNDTABLE  关注圆桌          :any:`StreamingJSON`
FOLLOW_TOPIC       关注话题          :any:`Topic`
LIKE_PIN           赞了分享          :any:`StreamingJSON`
JOIN_LIVE          参加 Live         :any:`Live`
PUBLISH_LIVE       举办 Live         :any:`Live`
VOTEUP_ANSWER      赞同回答          :any:`Answer`
VOTEUP_ARTICLE     赞同文章          :any:`Article`
VOTEUP_EBOOK       赞了电子书         :any:`StreamingJSON`
================= ================ ======================

收藏答案和收藏文章的 Target 属性是一个 dict，结构如下：

..  code-block:: javascript

    {
        'answer/article': <Answer object> or <Article object>,
        'collection': <Collection object>,
    }

``answer/article`` 表示被收藏的答案/文章，``collection`` 表示被收藏进的收藏夹，因为只有这两个动作有两个
操作对象，所以特殊处理了一下。

"""


def _verb_to_type(verb):
    type_str = _verb_to_type_map.get(verb, None)
    if type_str is None:
        raise UnimplementedException(
            'Unknown activity type: {0}'.format(verb)
        )
    return getattr(ActType, _verb_to_type_map[verb])


class Activity(object):
    def __new__(cls, data, session):
        if data['verb'] == 'MEMBER_FOLLOW_ROUNDTABLE':
            data['type'] = ActType.FOLLOW_ROUNDTABLE
            return StreamingJSON(data)
        elif data['verb'] == 'EBOOK_VOTE_UP':
            data['type'] = ActType.VOTEUP_EBOOK
            return StreamingJSON(data)
        else:
            return super(Activity, cls).__new__(cls)

    def __init__(self, data, session):
        """
        表示用户的一条动态。

        :any:`type <Activity.type>` 属性标识了动态的类型，其取值及意义请参见
        :any:`ActType`。

        :any:`target <Activity.target>` 属性表示这次动态操作的目标，根据
        `type` 的不同，这个属性的类型也不同。但是基本都是
        `type` 的最后一个单词表示的类型，请看下面的例子。

        ..  note:: 举例

            ..  code-block:: python

                from zhihu_oauth import ZhihuClient, ActType # 记得要导入 ActType

                client = ZhihuClient()

                # Client 登录过程省略

                me = client.me()

                for act in me.activities:
                    if act.type == ActType.CREATE_ANSWER:
                        print(act.target.question.title)

            上面这段代码只处理类型是创建答案的动态，此时 `act.target` 就是
            :any:`Answer` 类型，和 `CREATE_ANSWER` 的 `ANSWER` 对应。
        """
        self._data = data
        self._type = _verb_to_type(data['verb'])
        self._session = session
        self._get_target()

    @property
    @normal_attr()
    def action_text(self):
        return None

    @property
    @normal_attr()
    def created_time(self):
        """
        用户动态的时间戳。
        """
        return None

    @property
    def type(self):
        """
        动态的类型。

        ..  seealso:: :any:`ActType`
        """
        return self._type

    @property
    def target(self):
        """
        动态的操作目标。

        ..  seealso:: :any:`Activity.__init__`, :any:`ActType`
        """
        return self._target

    def _get_target(self):
        pos = self._type.rfind('_')
        if pos == -1:
            raise UnimplementedException('Unable to get class from type name')
        filename = self._type[pos + 1:].lower()
        class_name = filename.capitalize()
        module = importlib.import_module('.' + filename, 'zhihu_oauth.zhcls')
        cls = getattr(module, class_name)
        self._target = cls(self._data['target']['id'], self._data['target'],
                           self._session)

        # 对收藏答案类型的特殊处理
        if self._type in {ActType.COLLECT_ANSWER, ActType.COLLECT_ARTICLE}:
            from .collection import Collection
            obj = self._target
            collection = Collection(
                self._data['target']['collection']['id'],
                self._data['target']['collection'],
                self._session
            )
            self._target = {
                'collection': collection,
                class_name.lower(): obj,
            }



# coding=utf-8

from __future__ import unicode_literals

from .base import Base
from .other import other_obj
from .generator import generator_of
from .normal import normal_attr
from .urls import MESSAGES_URL

__all__ = ['Whisper']


class Whisper(Base):
    """
    唔，其实就是「和某人的所有消息」。

    为这个东西命名我想了半天……最后群里的一个小姐姐说叫 Whisper 吧，我觉得很可以诶~

    后来发现知乎接口里把这个叫做 Thread，嗯，不想改，我就是这么任性……
    """
    def _build_url(self):
        return ''

    def _get_data(self):
        pass

    def __init__(self, wid, cache, session):
        super(Whisper, self).__init__(wid, cache, session)

    # ----- simple info -----

    @property
    @normal_attr()
    def allow_reply(self):
        return None

    @property
    def id(self):
        return self._id

    @property
    @normal_attr()
    def snippet(self):
        """
        最后一次私信的摘要
        """
        return None

    @property
    @normal_attr()
    def updated_time(self):
        return None

    @property
    @normal_attr()
    def unread_count(self):
        return None

    @property
    @other_obj('people', 'participant')
    def who(self):
        """
        参与此私信会话的另一个知乎用户
        """
        return None

    @property
    @generator_of(MESSAGES_URL)
    def messages(self):
        return None



# coding=utf-8

from copy import deepcopy

from ..exception import UnimplementedException
from .utils import ConstValue, build_zhihu_obj_from_dict, SimpleEnum


__all__ = ['SearchResult', 'SearchResultSection', 'SearchType']

_search_type_t_map = {
    'GENERAL': 'general',
    'PEOPLE': 'people',
    'TOPIC': 'topic',
    'COLUMN': 'column',
    'LIVE': 'live',
    # 'EBOOK': 'publication',
    # EBOOK 是电子书搜索，目前电子书类还没写，暂时不开放。
}

SearchType = SimpleEnum(_search_type_t_map.keys())
"""
================= ================ ======================
常量名              说明              备注
================= ================ ======================
GENERAL            综合搜索         * 见下备注
PEOPLE             搜索用户
TOPIC              搜索话题
COLUMN             搜索专栏
LIVE               搜索 Live
EBOOK              搜索电子书        现在电子书类还没写，所以此常量暂时不能使用
================= ================ ======================

* 综合搜索是最复杂的一个，但是只有它能搜索到问题和答案。
  一般情况下综合搜索会返回几个除了问题和答案类型之外的 :any:`SearchResultSection`，
  然后返回 Answer，Article，Question 型的 :any:`SearchResult`，具体处理方法看后面文档吧。
"""


def search_type_to_t(search_type):
    return _search_type_t_map[search_type]


class SearchResult(object):
    _TYPE_KEY = ConstValue('type')
    _RESULT_INDICATOR = ConstValue('search_result')
    _RESULT_INDICATOR_TYPO = ConstValue('searach_result')

    _RESULT_HIGHLIGHT_KEY = ConstValue('highlight')
    _RESULT_OBJ_KEY = ConstValue('object')
    _HIGHLIGHT_TITLE_KEY = ConstValue('title')
    _HIGHLIGHT_DESC_KEY = ConstValue('description')

    def __init__(self, data, session):
        if data[self._TYPE_KEY] != self._RESULT_INDICATOR and \
                data[self._TYPE_KEY] != self._RESULT_INDICATOR_TYPO:
            raise ValueError("Must be a {} type dict, {} provided".format(
                self._RESULT_INDICATOR, data
            ))

        self._data = data
        self._session = session
        self._highlight = self._data.get(self._RESULT_HIGHLIGHT_KEY, {})

    @property
    def highlight_title(self):
        """
        标题，其中搜索关键词被高亮，是 HTML 格式的字符串，特殊字符被 escape 了，高亮的部分在 <em> 标签之间。
        print 出来的话不是很好读……
        """
        return self._highlight[self._HIGHLIGHT_TITLE_KEY] \
            if self._HIGHLIGHT_TITLE_KEY in self._highlight \
            else ''

    @property
    def highlight_desc(self):
        """
        description，搜索结果的内容。
        同 :any:`highlight_title`。
        """
        return self._highlight[self._HIGHLIGHT_DESC_KEY] \
            if self._HIGHLIGHT_DESC_KEY in self._highlight \
            else ''

    @property
    def obj(self):
        """
        搜索结果对应的知乎类对象，可能是各种类型，使用前需要自行判断，
        """
        obj = self._data[self._RESULT_OBJ_KEY]
        # 因为搜索结果里的用户属性，比如标题，名字等，都会被 <em> 标签标记成高亮，所以
        # 这里不使用他们当作 cache
        return build_zhihu_obj_from_dict(obj, self._session, use_cache=None)

    def raw_data(self):
        """
        返回搜索结果的原始数据的拷贝，是个 dict。
        """
        return deepcopy(self._data)


class SearchResultSection(object):
    _TYPE_KEY = ConstValue('type')
    _SECTION_INDICATOR = ConstValue('search_section')

    _SECTION_DATA_LIST_KEY = ConstValue('data_list')
    _SECTION_TYPE_KEY = ConstValue('section_type')
    _SECTION_HAS_MORE_KEY = ConstValue('has_more')

    def __init__(self, data, session):
        """
        :any:`SearchResultSection` 对象是可迭代的，`for xxx in results`
        一般会生成 :any:`SearchResult` 型数据。
        """
        if data[self._TYPE_KEY] != self._SECTION_INDICATOR:
            raise ValueError("Must be a {} type dict, {} provided".format(
                self._SECTION_INDICATOR, data
            ))

        self._data = data
        self._session = session

        self._index = 0
        self._len = len(self._data[self._SECTION_DATA_LIST_KEY])

    @property
    def type(self):
        """
        表示这一 Section 里的 :any:`SearchResult` 的知乎类对象是什么类型。
        :rtype: str|unicode
        """
        return self._data[self._SECTION_TYPE_KEY]

    @property
    def has_more(self):
        """
        如果用 type 类型进行搜索，能否得到更多结果。

        比如：::

            self.type == "people", self. has_more == True

        那么表示：::

            client.search('something', SearchType.PEOPLE)

        能获取到此 Section 的更多结果。

        :rtype: bool
        """
        return self._data[self._SECTION_HAS_MORE_KEY]

    def raw_data(self):
        """
        同 :any:`SearchResult.raw_data`
        """
        return deepcopy(self._data)

    def __iter__(self):
        self._index = 0
        return self

    def __len__(self):
        return self._len

    def __next__(self):
        try:
            obj = self[self._index]
        except IndexError:
            self._index = 0
            raise StopIteration
        self._index += 1
        return obj

    next = __next__

    def __getitem__(self, item):
        if not isinstance(item, int):
            raise TypeError('Need an int as index, not {0}'.format(type(item)))

        if item >= self._len:
            raise IndexError()

        data = self._data[self._SECTION_DATA_LIST_KEY][item]
        return data_to_section_or_result(data, self._session)


def data_to_section_or_result(data, session):
    try:
        return SearchResult(data, session)
    except ValueError:
        pass

    try:
        return SearchResultSection(data, session)
    except ValueError:
        raise UnimplementedException(
            "Unknown search result dict [{}]".format(data))



# coding=utf-8

from __future__ import unicode_literals

try:
    from json import JSONDecodeError as MyJSONDecodeError
except ImportError:
    MyJSONDecodeError = Exception
    """
    在 Py3 下是 json.JSONDecodeError。

    在 Py2 下是 BaseException，因为 Py2 的 json 模块没有提供解析异常。
    """

__all__ = [
    # warnings
    'ZhihuWarning',
    'IgnoreErrorDataWarning',
    'GetEmptyResponseWhenFetchData',
    'CantGetTicketsWarning',
    'CantGetTickets',
    # exceptions
    'ZhihuException',
    'UnexpectedResponseException',
    'GetDataErrorException',
    'NeedCaptchaException',
    'NeedLoginException',
    'IdMustBeIntException',
    'UnimplementedException',
    'MyJSONDecodeError',
]


class ZhihuException(Exception):
    pass


class UnexpectedResponseException(ZhihuException):
    def __init__(self, url, res, expect):
        """
        服务器回复了和预期格式不符的数据

        :param str|unicode url: 当前尝试访问的网址
        :param request.Response res: 服务器的回复
        :param str|unicode expect: 一个用来说明期望服务器回复的数据格式的字符串
        """
        self.url = url
        self.res = res
        self.expect = expect

    def __repr__(self):
        return 'Get an unexpected response when visit url ' \
               '[{self.url}], we expect [{self.expect}], ' \
               'but the response body is [{self.res.text}]'.format(self=self)

    __str__ = __repr__


class UnimplementedException(ZhihuException):
    def __init__(self, what):
        """
        处理当前遇到的情况的代码还未实现，只是开发的时候用于占位

        ..  note:: 一般用户不用管这个异常

        :param str|unicode what: 用来描述当前遇到的情况
        """
        self.what = what

    def __repr__(self):
        return 'Meet a unimplemented condition: {self.what}. ' \
               'Please send this error message to developer ' \
               'to get help.'.format(self=self)

    __str__ = __repr__


class GetDataErrorException(UnexpectedResponseException):
    def __init__(self, url, res, expect):
        """
        :class:`UnexpectedResponseException` 的子类，
        尝试获取服务器给出的错误信息。如果获取失败则显示父类的出错信息。

        ..  seealso:: :class:`UnexpectedResponseException`
        """
        super(GetDataErrorException, self).__init__(url, res, expect)
        try:
            self.reason = res.json()['error']['message']
        except (MyJSONDecodeError, KeyError):
            self.reason = None

    def __repr__(self):
        if self.reason:
            return 'A error happened when get data: {0}'.format(self.reason)
        else:
            base = super(GetDataErrorException, self).__repr__()
            return 'Unknown error! ' + base

    __str__ = __repr__


class NeedCaptchaException(ZhihuException):
    def __init__(self):
        """
        登录过程需要验证码
        """
        pass

    def __repr__(self):
        return 'Need a captcha to login, ' \
               'please catch this exception and ' \
               'use client.get_captcha() to get it.'

    __str__ = __repr__


class NeedLoginException(ZhihuException):
    def __init__(self, what):
        """
        使用某方法需要登录而当前客户端未登录

        :param str|unicode what: 当前试图调用的方法名
        """
        self.what = what

    def __repr__(self):
        return 'Need login to use the [{self.what}] method.'.format(self=self)

    __str__ = __repr__


class IdMustBeIntException(ZhihuException):
    def __init__(self, func):
        """
        获取对应的知乎类时，试图传递不是整数型的 ID

        :param function func: 当前试图调用的方法名
        """
        self.func = func.__name__

    def __repr__(self):
        return 'You must provide a integer id ' \
               'to use function: {self.func}'.format(self=self)

    __str__ = __repr__


class ZhihuWarning(UserWarning):
    def __init__(self, message, *args, **kwargs):
        super(ZhihuWarning, self).__init__(*args, **kwargs)
        self._message = message

    def __str__(self):
        return str(self._message)

    __repr__ = __str__


class IgnoreErrorDataWarning(ZhihuWarning):
    def __init__(self, message, *args, **kwargs):
        super(IgnoreErrorDataWarning, self).__init__(message, *args, **kwargs)


GetEmptyResponseWhenFetchData = IgnoreErrorDataWarning(
    '试图获取下一项时，服务器返回了空数据。'
    '如果您是在获取某用户的粉丝，那么您可能遇到了知乎 5020 限制。'
    '虽然不知道为什么，但是好像知乎限制 API 只能访问前 5020 个粉丝，'
    '我也很为难，但是这是知乎做的限制，突破不了呀。'
    '目前在遇到这个问题时只能当作获取完处理了。'
)


class CantGetTicketsWarning(ZhihuWarning):
    def __init__(self, message, *args, **kwargs):
        super(CantGetTicketsWarning, self).__init__(message, *args, **kwargs)

CantGetTickets = CantGetTicketsWarning(
    '只能获取未参与过（也即 live.role == \'visitor\'）的 Live 的票价信息，'
    '对于其他情况此接口不会返回任何东西。这不会造成 for in 循环报错，'
    '但如果你不想看到这个警告，请……请在调用 tickets 属性前自行判断一下，'
    '或者偷懒强行过滤掉这个 warning ：）'
)



# coding=utf-8

from __future__ import unicode_literals

from .base import Base
from .generator import generator_of
from .normal import normal_attr
from .streaming import streaming
from .urls import (
    PEOPLE_ACTIVITIES_URL,
    PEOPLE_ANSWERS_URL,
    PEOPLE_ARTICLES_URL,
    PEOPLE_COLLECTIONS_URL,
    PEOPLE_COLUMNS_URL,
    PEOPLE_DETAIL_URL,
    PEOPLE_FOLLOWERS_URL,
    PEOPLE_FOLLOWING_COLUMNS_URL,
    PEOPLE_FOLLOWING_QUESTIONS_URL,
    PEOPLE_FOLLOWING_TOPICS_URL,
    PEOPLE_FOLLOWINGS_URL,
    PEOPLE_LIKED_LIVES_URL,
    PEOPLE_LIVES_URL,
    PEOPLE_PINS_URL,
    PEOPLE_QUESTIONS_URL,
)
from .utils import ConstValue
from ..exception import GetDataErrorException

__all__ = ['ANONYMOUS', 'Badge', 'People']


class _Anonymous(object):
    def __init__(self):
        self.id = 0
        self.name = '匿名用户'

    def __getattr__(self, _):
        # 匿名用户除了姓名和 ID 以外所有属性均为 None
        return None


ANONYMOUS = _Anonymous()
"""
.. role:: py_code(code)
   :language: python

统一的匿名用户对象，可以使用 :py_code:`if people is ANONYMOUS:` 判断是否是匿名用户
"""


class People(Base):
    def __new__(cls, pid, cache, session):
        if pid == '0':
            return ANONYMOUS
        else:
            return super(People, cls).__new__(cls)

    def __init__(self, pid, cache, session):
        self._over_e = None
        super(People, self).__init__(pid, cache, session)

    def _build_url(self):
        return PEOPLE_DETAIL_URL.format(self.id)

    @property
    def over(self):
        """
        尝试获取用户信息，如果出错返回 True，没出错返回 False。

        一般来说出错的情况只有「被知乎反屏蔽系统限制」……所以这函数起名叫 over =。=

        调用结果如果是 True，则可使用 :any:`over_reason` 函数获取原因。

        .. note:: 例子

            ..  code-block:: python

                for follower in me.followers:
                    if follower.over:
                        print(follower.over_reason)
                        continue

                    print(follower.name)
                    # ... process follower data

        ..  note:: 也可以不用此函数，用 ``try...catch`` 来处理也行

            ..  code-block:: python

                from zhihu_oauth import ZhihuClient, GetDataErrorException

                for follower in me.followers:
                    try:
                        # get and process user data
                    except GetDataErrorException as e:
                        print('Get data error', e.reason)

        :return: 是否被限制
        :rtype: bool
        """
        if self._over_e is not None:
            return True
        try:
            self._get_data()
            return False
        except GetDataErrorException as e:
            self._over_e = e
            return True

    @property
    def over_reason(self):
        """
        获取无法得到用户信息的原因。

        ..  warning::

            此方法只能在 :any:`over` 方法调用结果为 True 之后才能调用。

        :rtype: str
        """
        if self._over_e is None:
            return None
        return self._over_e.reason if self._over_e.reason else str(self._over_e)

    # ---------- simple info ---------

    @property
    @normal_attr()
    def answer_count(self):
        return None

    @property
    @normal_attr('articles_count')
    def article_count(self):
        return None

    @property
    def articles_count(self):
        return self.article_count

    @property
    @normal_attr()
    def avatar_url(self):
        return None

    @property
    @streaming('badge', use_cache=False)
    def _badge_data(self):
        return []

    @property
    @streaming('org_detail', use_cache=False)
    def _org_data(self):
        return {}

    @property
    def badge(self):
        """
        用户的徽章信息，获取到的是 :any:`Badge` 对象，使用方法请看 :any:`Badge` 类的文档。

        目前包括「个人认证」，「话题优秀回答者」，「已认证机构」三种，
        我也不知道有没有更多的呀，如果发现这个功能有 Bug 或者还有更多类型的徽章，
        请到 Github 里提 Issue 或者直接联系我。

        :rtype: Badge
        """
        return Badge(self._badge_data, self._org_data, self._session)

    @property
    @streaming('badge')
    def _cached_badge_data(self):
        return []

    @property
    @streaming('org_detail')
    def _cached_org_data(self):
        return {}

    @property
    def cached_badge(self):
        return Badge(
            self._cached_badge_data,
            self._cached_org_data,
            self._session
        )

    @property
    @streaming()
    def business(self):
        """
        用户所在行业。

        常见返回值：

        ..  code-block:: python

            {
                'introduction': '',
                'id': '19619368',
                'url': 'https://api.zhihu.com/topics/19619368',
                'type': 'topic',
                'avatar_url': 'http://pic1.zhimg.com/e82bab09c_s.jpg',
                'name': '计算机软件',
                'excerpt': '',
            }

        使用属性时必须先判断是否有效，如

        ..  code-block:: python

            if 'people.business:
                data = people.business.name

        """
        return {}

    @property
    @normal_attr('favorited_count')
    def collected_count(self):
        """
        被收藏次数。
        """
        return None

    @property
    @normal_attr('favorite_count')
    def collection_count(self):
        """
        收藏夹数量。
        """
        return None

    @property
    @normal_attr('columns_count')
    def column_count(self):
        return None

    @property
    def columns_count(self):
        return self.column_count

    @property
    @normal_attr()
    def description(self):
        return None

    @property
    @streaming()
    def educations(self):
        """
        教育信息。

        常见返回值：

        ..  code-block:: python

            [
                {
                    'major': {
                        'introduction': '计算机专业。<br>大众认为会是唯一会“修电脑”的专业。',
                        'id': '19639658',
                        'url': 'https://api.zhihu.com/topics/19639658',
                        'type': 'topic',
                        'avatar_url': 'http://pic2.zhimg.com/7e2fe4615_s.jpg',
                        'name': '计算机科学与技术',
                        'excerpt': '计算机专业。大众认为会是唯一会“修电脑”的专业。',
                    },
                    'school': {
                        'introduction': '',
                        'id': '1234567',
                        'url': 'https://api.zhihu.com/topics/1234567',
                        'type': 'topic',
                        'avatar_url': 'http://pic4.zhimg.com/8e6y3xd47_s.jpg',
                        'name': 'XX 大学',
                        'excerpt': '',
                    },
                },
            ]

        使用属性时必须先判断存不存在，如:

        ..  code-block:: python

            for education in people.educations:
                if 'school' in education:
                    data += education.school.name
                if 'major' in education:
                    data += education.major.name'
        """
        return []

    @property
    @streaming()
    def employments(self):
        """
        职业信息。

        常见返回值：

        ..  code-block:: python

            [
                {
                    'job': {
                        'introduction': '',
                        'url': 'https://api.zhihu.com/topics/19551336',
                        'avatar_url': 'http://pic3.zhimg.com/4eac47b76_s.jpg',
                        'excerpt': '',
                        'type': 'topic',
                        'name': '测试',
                        'id': '19551336',
                    },
                    'company': {
                        'excerpt': '',
                        'url': '',
                        'avatar_url': 'http://pic1.zhimg.com/e82bab09c_s.jpg',
                        'introduction': '',
                        'type': 'topic',
                        'name': 'Gayhub',
                        'experience ': '',
                        'id': '',
                    },
                },
            ],

        使用属性时必须先判断存不存在，如:

        ..  code-block:: python

            for employment in people.employments:
                if 'company' in employment:
                    data += employment.company.name
                if 'job' in employment:
                    data += employment.job.name'
        """
        return []

    @property
    def favorite_count(self):
        return self.collection_count

    @property
    def favorited_count(self):
        return self.collected_count

    @property
    @normal_attr()
    def follower_count(self):
        return None

    @property
    @normal_attr('following_columns_count')
    def following_column_count(self):
        return None

    @property
    @normal_attr()
    def following_count(self):
        return None

    @property
    @normal_attr()
    def following_question_count(self):
        return None

    @property
    @normal_attr()
    def following_topic_count(self):
        return None

    @property
    @normal_attr()
    def gender(self):
        """
        性别。

        =======  ==========
        值(int)  说明
        =======  ==========
        0        女
        1        男
        -1       未填
        =======  ==========

        ..  warning:: 关于未填性别

            目前版本的知乎 API 貌似未填性别的用户这个值也会变成 1

            已经向知乎反馈了 Bug，但是不知道什么时候才会修。
        """
        return None

    @property
    @normal_attr()
    def headline(self):
        """
        就是那个显示在名字后面的，和签名类似的东西。
        """
        return None

    @property
    @normal_attr()
    def hosted_live_count(self):
        return None

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr()
    def independent_articles_count(self):
        return None

    @property
    @normal_attr()
    def is_bind_sina(self):
        return None

    @property
    @normal_attr()
    def is_blocking(self):
        return None

    @property
    @normal_attr('is_followed')
    def is_follower(self):
        """
        :return: 此人是否关注当前登录用户。对于当前登录用户值为 False。
        :rtype: bool
        """
        return None

    @property
    @normal_attr()
    def is_following(self):
        """
        :return: 当前登录用户是否已关注此人。对于当前登录用户值为 False。
        :rtype: bool
        """
        return None

    @property
    @normal_attr()
    def live_count(self):
        """
        包括参与的和组织的 Live。
        """
        return None

    @property
    @streaming()
    def locations(self):
        """
        常见返回值。

        ..  code-block:: python

            [
                {
                    'introduction': '天津，简称津，地处华北平原，balabala,
                    'url': 'https://api.zhihu.com/topics/19577238',
                    'avatar_url': 'http://pic4.zhimg.com/acad405e7_s.jpg',
                    'excerpt': '天津，简称津，地处华北平原 balabalabala',
                    'type': 'topic',
                    'name': '天津',
                    'id': '19577238',
                },
            ],

        使用属性时基本不用判断存不存在，如:

        ..  code-block:: python

            for location in people.locations:
                data += location.name
        """
        return []

    @property
    @normal_attr()
    def name(self):
        return None

    @property
    @normal_attr()
    def participated_live_count(self):
        return None

    @property
    @normal_attr('pins_count')
    def pin_count(self):
        return None

    @property
    @normal_attr()
    def question_count(self):
        return None

    @property
    @normal_attr()
    def shared_count(self):
        return None

    @property
    @normal_attr()
    def sina_weibo_name(self):
        return None

    @property
    @normal_attr()
    def sina_weibo_url(self):
        return None

    @property
    @normal_attr()
    def thanked_count(self):
        return None

    @property
    @normal_attr()
    def voteup_count(self):
        return None

    # ---------- generators ---------

    @property
    @generator_of(PEOPLE_ACTIVITIES_URL, 'activity')
    def activities(self):
        return None

    @property
    @generator_of(PEOPLE_ANSWERS_URL)
    def answers(self):
        return None

    @property
    @generator_of(PEOPLE_ARTICLES_URL)
    def articles(self):
        return None

    @property
    @generator_of(PEOPLE_COLLECTIONS_URL)
    def collections(self):
        return None

    @property
    @generator_of(PEOPLE_COLUMNS_URL)
    def columns(self):
        return None

    @property
    @generator_of(PEOPLE_FOLLOWERS_URL, 'people')
    def followers(self):
        """
        貌似知乎 API 有个限制，只允许获取前 5020 个粉丝，这好烦阿……
        """
        return None

    @property
    @generator_of(PEOPLE_FOLLOWING_COLUMNS_URL, 'column')
    def following_columns(self):
        return None

    @property
    @generator_of(PEOPLE_FOLLOWING_QUESTIONS_URL, 'question')
    def following_questions(self):
        return None

    @property
    @generator_of(PEOPLE_FOLLOWING_TOPICS_URL, 'topic')
    def following_topics(self):
        return None

    @property
    @generator_of(PEOPLE_FOLLOWINGS_URL, 'people')
    def followings(self):
        return None

    @property
    @generator_of(PEOPLE_LIVES_URL)
    def lives(self):
        """
        举办和参加的 Live
        """
        return None

    @property
    @generator_of(PEOPLE_LIKED_LIVES_URL, 'live')
    def liked_lives(self):
        """
        喜爱的 Live

        ..  warning:: 此接口未测试，不保证可用性。
        """
        return None

    @property
    @generator_of(PEOPLE_PINS_URL)
    def pins(self):
        return None

    @property
    @generator_of(PEOPLE_QUESTIONS_URL)
    def questions(self):
        return None


class Badge(object):
    IDENTITY = ConstValue('identity')
    BEST_ANSWERER = ConstValue('best_answerer')
    ORGANIZATION = ConstValue('organization')

    def __init__(self, badge, org, session):
        self._badge_data = badge
        self._org_data = org
        self._session = session

    @property
    def has_badge(self):
        """
        是否有徽章。

        有「个人认证」，「话题最佳回答者」，「已认证机构」 徽章中的**任何一个**就返回 ``True``。
        没有徽章返回 ``False``。
        """
        return bool(self._badge_data or self._org_data)

    @property
    def has_identity(self):
        """
        是否有「个人认证」徽章。
        """
        return bool(self._badge_data) and any(
            [x.type == self.IDENTITY for x in self._badge_data]
        )

    @property
    def is_best_answerer(self):
        """
        是否有「最佳回答者」徽章。
        """
        return bool(self._badge_data) and any(
            [x.type == self.BEST_ANSWERER for x in self._badge_data]
        )

    @property
    def is_organization(self):
        """
        是否为「已认证机构」。
        """
        return bool(self._org_data)

    @property
    def identity(self):
        """
        「个人认证」的描述。

        如果没有个人认证返回 ``None``。
        """
        for badge in self._badge_data:
            if badge.type == self.IDENTITY:
                return badge.description
        return None

    @property
    def topics(self):
        """
        是哪些话题的最佳回答者，返回的是话题的迭代器。

        可以这样用：

        ..  code-block:: python

            for topic in people.badge.topics:
                print(topic.name)


        :rtype: Iterable[Topic]
        """
        from .topic import Topic
        for badge in self._badge_data:
            if badge.type == self.BEST_ANSWERER:
                for topic_data in badge.topics:
                    yield Topic(topic_data.id, topic_data.raw_data(),
                                self._session)
                break

    @property
    def org_name(self):
        """
        认证的机构名称，绝大多数情况是公司名称。

        如果不是认证机构返回 ``None``。
        """
        if not self.is_organization:
            return None
        return self._org_data.organization_name

    @property
    def org_home_page(self):
        """
        认证的机构主页。

        如果不是认证机构返回 ``None``。
        """
        if not self.is_organization:
            return None
        return self._org_data.home_page

    @property
    def org_industry(self):
        """
        认证的机构所在行业。

        如果不是认证机构返回 ``None``。
        """
        if not self.is_organization:
            return None
        return self._org_data.industry



from .client_test_base import ZhihuClientClassTest

PEOPLE_SLUG = 'giantchen'


class TestPeopleBadgeNumber(ZhihuClientClassTest):
    def test_badge_topics_number(self):
        self.assertEqual(
            len(list(self.client.people(PEOPLE_SLUG).badge.topics)), 3
        )

    def test_people_has_badge(self):
        self.assertTrue(self.client.people(PEOPLE_SLUG).badge.has_badge)

    def test_people_has_identity(self):
        self.assertFalse(self.client.people(PEOPLE_SLUG).badge.has_identity)

    def test_people_is_best_answerer_or_not(self):
        self.assertTrue(self.client.people(PEOPLE_SLUG).badge.is_best_answerer)

    def test_people_identify_information(self):
        self.assertIsNone(self.client.people(PEOPLE_SLUG).badge.identity)



# coding=utf-8

from __future__ import unicode_literals

import re
import requests.adapters

ADAPTER_WITH_RETRY = requests.adapters.HTTPAdapter(
    max_retries=requests.adapters.Retry(
        total=10,
        status_forcelist=[403, 408, 500, 502]
    )
)

DEFAULT_CAPTCHA_FILENAME = 'captcha.gif'
"""
请求验证码后储存文件名的默认值，现在的值是当前目录下的 captcha.gif。

仅在 :meth:`.ZhihuClient.login_in_terminal` 中被使用。
"""



# coding=utf-8

from .activity import Activity, ActType
from .answer import Answer
from .article import Article
from .collection import Collection
from .column import Column
from .comment import Comment
from .live import Live, LiveBadge, LiveTag, LiveTicket
from .me import Me
from .message import Message
from .people import Badge, People, ANONYMOUS
from .pin import Pin, PinContent, PCType
from .question import Question
from .search import SearchResultSection, SearchResult, SearchType
from .topic import Topic
from .whisper import Whisper

__all__ = [
    'Activity', 'ActType', 'Answer', 'Article', 'ANONYMOUS', 'Badge',
    'Collection', 'Column', 'Comment', 'Live', 'LiveBadge', 'LiveTag',
    'LiveTicket', 'Me', 'Message', 'People', 'Pin', 'PinContent', 'PCType',
    'Question', 'SearchResult', 'SearchResultSection', 'SearchType',
    'Topic', 'Whisper',
]



# coding=utf-8

from .client import ZhihuClient
from .exception import (
    ZhihuWarning, IgnoreErrorDataWarning, CantGetTicketsWarning,
    ZhihuException, UnexpectedResponseException, GetDataErrorException,
    NeedCaptchaException, NeedLoginException, IdMustBeIntException,
    UnimplementedException,
)
from .helpers import act2str, ActivityFormatter, shield, SHIELD_ACTION, ts2str
from .zhcls import (
    Activity, ActType, Answer, ANONYMOUS, Article,Badge, Comment, Collection,
    Column, Comment, Live, LiveBadge, LiveTag, LiveTicket, Me, Message, People,
    Pin, PinContent, PCType, Question, SearchResult, SearchResultSection,
    SearchType, Topic, Whisper,
)

__all__ = [
    'Activity', 'ActType', 'Answer', 'ANONYMOUS', 'Article',
    'Collection', 'Column', 'Comment', 'Live', 'LiveBadge', 'LiveTag',
    'LiveTicket', 'Me', 'Message', 'People', 'Question', 'SearchResult',
    'SearchResultSection', 'SearchType', 'Topic', 'Whisper', 'ZhihuClient',
    'ZhihuException', 'ZhihuWarning',
    'NeedCaptchaException', 'UnexpectedResponseException',
    'GetDataErrorException',
    'act2str', 'shield', 'ts2str', 'ActivityFormatter', 'SHIELD_ACTION'
]

__version__ = '0.0.36'

# TODO: remove all magic number and magic string



# coding=utf-8

from __future__ import unicode_literals

import itertools

from .base import Base
from .generator import generator_of
from .normal import normal_attr
from .other import other_obj
from .streaming import streaming, StreamingJSON
from .urls import (
    PIN_COMMENTS_URL,
    PIN_DETAIL_URL,
    PIN_VOTERS_URL,
    ZHIHU_WEB_ROOT,
)
from .utils import (
    get_class_from_name,
    zhihu_obj_url_parse,
    SimpleEnum,
)

from ..exception import UnimplementedException

__all__ = ['Pin', 'PinContent', 'PCType']


class Pin(Base):

    def __init__(self, pid, cache, session):
        super(Pin, self).__init__(pid, cache, session)

    def _build_url(self):
        return PIN_DETAIL_URL.format(self.id)

    # ----- simple attrs -----

    @property
    @other_obj('people')
    def author(self):
        return None

    @property
    @normal_attr()
    def id(self):
        return self._id

    @property
    @normal_attr()
    def comment_count(self):
        return None

    @property
    @normal_attr()
    def comment_permission(self):
        return None

    @property
    @normal_attr('created')
    def created_time(self):
        return None

    @property
    @normal_attr()
    def excerpt_title(self):
        """
        简要标题文字
        """
        return None

    @property
    @normal_attr()
    def like_count(self):
        return None

    @property
    @normal_attr('updated')
    def updated_time(self):
        return None

    @property
    @streaming('content')
    def _contents(self):
        return []

    @property
    def contents(self):
        """
        因为一个分享里含有多个内容，所以 contents 属性是迭代器，需要这样使用：

        ..  code-block:: Python

            for act in someone.activities.filter(ActType.LINK_PIN):
                pin = act.target
                for pc in pin.contents:
                    # do something with pc

        迭代器返回 :any:`PinContent` 类型的对象，具体用法参看
        :any:`_pin_content_type_map` 和 any:`PinContent`。
        """
        for data in self._contents:
            yield PinContent(data, self._session)

    @property
    @normal_attr()
    def content_html(self):
        return None

    # ----- generators -----

    @property
    @generator_of(PIN_COMMENTS_URL)
    def comments(self):
        return None

    @property
    @generator_of(PIN_VOTERS_URL, 'people')
    def voters(self):
        return None


_pin_content_type_map = {
    'text': 'TEXT',
    'image': 'IMAGE',
    'link': 'LINK',
    'quote': 'QUOTE',
}
"""
:any:`PinContent` 的类型，每种类型能够使用的属性都不一样，具体请看下表：

+--------+---------------+------------------------------------+
|  类型  |      属性     |               含义                 |
+========+===============+====================================+
| TEXT   |      text     | 用户对分享的评论文字               |
+--------+---------------+------------------------------------+
|        |      src      | 图片地址                           |
| IMAGE  +---------------+------------------------------------+
|        | width, height | 图片宽高                           |
+--------+---------------+------------------------------------+
|        |     title     | 网页的标题                         |
|        +---------------+------------------------------------+
|  LINK  |      url      | 网页地址                           |
|        +---------------+------------------------------------+
|        |   image_url   | 网页的 ICON 地址                   |
+--------+---------------+------------------------------------+
|  QUOTE |     quote     | 引用的文字                         |
+--------+---------------+------------------------------------+
|        |     type      | 类型，可用 is PcType.TEXT 判断     |
|        +---------------+------------------------------------+
|        |    subtype    | 表示内容中包含的知乎对象类型，     |
|        |               | 参见 :any:`_subtype_zhihu_obj_map` |
|  通用  +---------------+------------------------------------+
|        |       obj     | 分享内容是知乎相关链接的话         |
|        |               | 可使用此属性获取到对应知乎类对象   |
|        +---------------+------------------------------------+
|        |               | 获取主要的数据                     |
|        |     content   | TEXT -> text, LINK -> url          |
|        |               | IMAGE -> src, QUOTE -> quote       |
+--------+---------------+------------------------------------+
"""

_subtype_zhihu_obj_map = {
    '问题': 'QUESTION',
    '回答': 'ANSWER',
    '专栏': 'COLUMN',
    '文章': 'ARTICLE',
    '收藏夹': 'COLLECTION',
    '话题': 'TOPIC',
    '个人主页': 'PEOPLE',
    'Live': 'LIVE',
    # ----- no zhihu object -----
    'none': 'NONE',
}
"""
:any:`PinContent.obj_type` 的取值枚举。

使用方法大概如下：

..  code-block:: Python

    for act in someone.activities.filter(ActType.LIKE_PIN):
        pin = act.target
        for content in pin.contents:
            if content.obj_type not is PCType.NONE:
                print(content.obj)
"""

_type_zhihu = 'zhihu'

PCType = SimpleEnum(
    itertools.chain(
        _pin_content_type_map.values(),
        _subtype_zhihu_obj_map.values(),
    )
)
"""
参见 :any:`_pin_content_type_map` 和 :any:`_subtype_zhihu_obj_map`。
"""

_type_content_method_map = {
    PCType.TEXT: 'text',
    PCType.IMAGE: 'src',
    PCType.QUOTE: 'quote',
    PCType.LINK: 'url'
}
"""
从 :any:`PinContent` 的类型对应到 :any:`PinContent.content` 属性需要使用的方法名
"""


class PinContent(object):
    """
    PinContent 是表示分享的内容的类，因为其类型很多，所以单独成类，设计理念和
    :any:`Activity` 类似。

    每一个 :any:`Pin` 可能含有多个 :any:`PinContent`，
    一般第一个是用户自对自己分享的东西的评论，余下的是 TA 分享的东西。

    :any:`PinContent` 主要有四个类型，具体请看 :any:`_pin_content_type_map`。

    使用方法如下：

    ..  code-block:: Python

        for act in me.activities.filter({ActType.LIKE_PIN, ActType.CREATE_PIN}):
            for content in act.target.contents:
                print(content.type, content.obj_type, end=' ')
                if content.type is PCType.TEXT:
                    print(content.text)
                elif content.type is PCType.QUOTE:
                    print(content.quote)
                elif content.type is PCType.LINK:
                    print(content.title, content.url, content.image_url)
                elif content.type is PCType.IMAGE:
                    print(content.src, content.width, content.height)
            print('-' * 20)

    或者使用 :any:`PinContent.content` 获取对于每种类型的主要内容。
    参见 :any:`_type_content_method_map`
    """
    def __init__(self, data, session):
        assert isinstance(data, StreamingJSON)
        self._data = data
        self._session = session

    # ----- Text type ----

    @property
    def text(self):
        assert self.type is PCType.TEXT
        return self._data.content

    # ----- Image type -----

    @property
    def src(self):
        assert self.type is PCType.IMAGE
        return self._data.url

    @property
    def width(self):
        assert self.type is PCType.IMAGE
        return self._data.width

    @property
    def height(self):
        assert self.type is PCType.IMAGE
        return self._data.height

    # ----- Normal link type -----

    @property
    def title(self):
        assert self.type is PCType.LINK
        return self._data.title

    @property
    def url(self):
        assert self.type is PCType.LINK or self.type is PCType.QUOTE
        return self._data.url

    @property
    def image_url(self):
        assert self.type is PCType.LINK
        return self._data.image_url

    # ----- Quote -----

    @property
    def quote(self):
        assert self.type is PCType.QUOTE
        return self._data.content

    # ------ Common -----

    @property
    def type(self):
        if self._data.type not in _pin_content_type_map:
            raise UnimplementedException(
                'Unknown pin content type [{}]'.format(self._data.type)
            )

        return getattr(PCType, _pin_content_type_map[self._data.type])

    @property
    def obj_type(self):
        if self.type is PCType.LINK and self._data.link_type == _type_zhihu:
            subtype = self._data.subtype
            if subtype not in _subtype_zhihu_obj_map:
                raise UnimplementedException(
                    'Unknown pin link obj_type [{}]'.format(subtype)
                )
            return getattr(PCType, _subtype_zhihu_obj_map[subtype])

        if self.type is PCType.QUOTE:
            obj_id, obj_type, _ = zhihu_obj_url_parse(self._data.url)
            if obj_id:
                return getattr(PCType, obj_type.upper())
            elif self._data.url.startswith(ZHIHU_WEB_ROOT):
                raise UnimplementedException(
                    'Unknown pin quote obj_type, data [{}]'.format
                    (self._data)
                )

        return PCType.NONE

    @property
    def content(self):
        return getattr(self, _type_content_method_map[self.type])

    @property
    def obj(self):
        assert self.obj_type is not PCType.NONE

        if self.type is PCType.LINK:
            obj_id = self._data.token
        else:
            assert self.type is PCType.QUOTE
            obj_id, _, _ = zhihu_obj_url_parse(self.url)

        cls = get_class_from_name(self.obj_type)
        return cls(obj_id, None, self._session)
